#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from queue import Queue
import asyncio as aio
from typing import Callable, Tuple

# Class Type: an async class
# Class Purpose:
#   to have a window-based method dealing with asyncio tasks
class AsyncWindow:
    class TaskItem:
        __slots__ = ('function','args')
        def __init__(self, function:Callable, args:Tuple) -> None:
            self.function, self.args = function, args
    def __init__(self, windowSize:int) -> None:
        self.maxWindow, self.currtasks, self.tasks = windowSize, [], Queue()
    def addTask(self, function:Callable, args:Tuple) -> None:
        task:TaskItem = self.TaskItem(function, args)
        self.tasks.put(task)
        self._processQueue()
    async def gather(self) -> None:
        while self.tasks.qsize() != 0 or self.currtasks:
            await aio.sleep(0.001)
    def getNumTasks(self) -> int:
        return self.tasks.qsize()
    def getWindowSize(self) -> int:
        return self.maxWindow
    async def _task(self, task:TaskItem) -> None:
        await task.function(*task.args)
        self.currtasks.remove(task)
        self._processQueue()
    def _processQueue(self) -> None:
        while self.tasks.qsize()>0 and len(self.currtasks)<self.maxWindow:
            task = self.tasks.get_nowait()
            self.currtasks.append(task)
            aio.create_task(self._task(task))


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

from __future__ import annotations

# Basic Libraries
from struct import unpack_from
from typing import List, Optional
# NDN Imports
from ndn.encoding import Component, get_tl_num_size, write_tl_num, parse_tl_num, pack_uint_bytes
# Custom Imports
from .tlv import SVSyncTlvTypes

# Class Type: a struct
# Class Purpose:
#   to hold info about a singular node within the vector.
class StateVectorEntry:
    __slots__ = ('nid','seqno')
    def __init__(self, nid:str, seqno:int) -> None:
        self.nid, self.seqno = nid, seqno
    def encode(self) -> bytearray:
        bseqno, bnid = pack_uint_bytes(self.seqno), self.nid.encode()
        # nid
        size = len(bnid) + get_tl_num_size(len(bnid)) + get_tl_num_size(SVSyncTlvTypes.VECTOR_ENTRY.value)
        temp1 = bytearray(size)
        pos = write_tl_num(SVSyncTlvTypes.VECTOR_ENTRY.value, temp1)
        pos += write_tl_num(len(bnid), temp1, pos)
        temp1[pos:pos + len(bnid)] = bnid
        # seqno
        size = len(bseqno) + get_tl_num_size(len(bseqno)) + get_tl_num_size(SVSyncTlvTypes.SEQNO.value)
        temp2 = bytearray(size)
        pos = write_tl_num(SVSyncTlvTypes.SEQNO.value, temp2)
        pos += write_tl_num(len(bseqno), temp2, pos)
        temp2[pos:pos + len(bseqno)] = bseqno
        return temp1+temp2

# Class Type: an custom tlv model class
# Class Purpose:
#   to contain all the info of a state vector.
class StateVectorModel:
    def __init__(self, value:List[StateVectorEntry]) -> None:
        self.value:List[StateVectorEntry] = value
    def encode(self) -> bytearray:
        length, component_wires = 0, []
        for v in self.value:
            ba = v.encode()
            length += len(ba)
            component_wires.append(ba)
        buf_len = length + get_tl_num_size(length) + get_tl_num_size(SVSyncTlvTypes.VECTOR.value)
        ret = bytearray(buf_len)
        pos = write_tl_num(SVSyncTlvTypes.VECTOR.value, ret)
        pos += write_tl_num(length, ret, pos)
        for w in component_wires:
            ret[pos:pos + len(w)] = w
            pos += len(w)
        return ret
    @staticmethod
    def parse(buf:Component) -> Optional[StateVectorModel]:
        # Verify the Type
        typ, pos = parse_tl_num(buf)
        if typ != SVSyncTlvTypes.VECTOR.value:
            return None
        # Check the length
        length, l = parse_tl_num(buf, pos)
        pos += l
        if pos + length != len(buf):
            return None
        # Decode components
        ret:StateVectorModel = StateVectorModel([])
        ret.value = []
        while pos < len(buf):
            # Entry ID
            typ, l = parse_tl_num(buf, pos)
            pos += l
            if typ != SVSyncTlvTypes.VECTOR_ENTRY.value:
                return None
            length, l = parse_tl_num(buf, pos)
            pos += l
            entry = buf[pos:pos + length]
            pos += length
            # Seqno
            typ, l = parse_tl_num(buf, pos)
            pos += l
            if typ != SVSyncTlvTypes.SEQNO.value:
                return None
            length, l = parse_tl_num(buf, pos)
            pos += l
            if length == 1:
                seqno = unpack_from('!B', buf, pos)[0]
            elif length == 2:
                seqno = unpack_from('!H', buf, pos)[0]
            elif length == 4:
                seqno = unpack_from('!I', buf, pos)[0]
            elif length == 8:
                seqno = unpack_from('!Q', buf, pos)[0]
            else:
                return None
            pos += length
            # Append the component
            ret.value.append(StateVectorEntry(bytes(entry).decode(), seqno))
        return ret

# Class Type: a class
# Class Purpose:
#   to allow an easier time to interact with the StateVectorModel class.
class StateVector:
    def __init__(self, component:Component=None) -> None:
        self.vector, self.wire = StateVectorModel([]), None
        if component:
            temp_vector:Optional[StateVectorModel] = StateVectorModel.parse(component)
            if temp_vector != None:
                self.vector = temp_vector
    def set(self, nid:str, seqno:int, oldData:bool=False) -> None:
        index, self.wire = self.index(nid), None
        if index == None:
            svc:StateVectorEntry = StateVectorEntry(nid, seqno)
            if not oldData:
                self.vector.value.insert(0, svc)
            else:
                self.vector.value.append(svc)
        else:
            self.vector.value[index].seqno = seqno
            if not oldData:
                self.vector.value.insert(0, self.vector.value.pop(index))
    def get(self, nid:str) -> Optional[int]:
        for i in self.vector.value:
            if i.nid == nid:
                return i.seqno
        return None
    def has(self, nid:str) -> bool:
        return False if self.get(nid) == None else True
    def index(self, nid:str) -> Optional[int]:
        for index, value in enumerate(self.vector.value):
            if value.nid == nid:
                return index
        return None
    def to_str(self) -> str:
        return " ".join([( i.nid + ":" + str(i.seqno) ) for i in self.vector.value])
    def encode(self) -> bytes:
        if self.wire == None:
            self.wire = self.vector.encode()
        return self.wire
    def keys(self) -> List[str]:
        return [i.nid for i in self.vector.value]
    def list(self) -> List[StateVectorEntry]:
        return self.vector.value
    def to_component(self) -> Component:
        return self.encode()
    def length(self) -> int:
        return len(self.encode())
    def partition(self, start:int, end:int) -> Component:
        return StateVectorModel(self.vector.value[start:end]).encode()
    def entry(self, index:int) -> Optional[StateVectorEntry]:
        try:
            return self.vector.value[index]
        except (IndexError, ValueError):
            return None
    def total(self) -> int:
        return sum(i.seqno for i in self.vector.value)
    def entry_lengths(self) -> List[int]:
        return [len(i.encode()) for i in self.vector.value]


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from typing import Optional, Callable, Tuple
# NDN Imports
from ndn.app import NDNApp
from ndn.encoding import Name, MetaInfo, InterestParam, BinaryStr, FormalName, SignatureType, make_data, parse_data
from ndn.types import InterestNack, InterestTimeout, InterestCanceled, ValidationFailure
from ndn.storage import Storage, MemoryStorage
# Custom Imports
from .core import SVSyncCore
from .logger import SVSyncLogger
from .security import SecurityOptions, SigningInfo, ValidatingInfo

# Class Type: an abstract API class
# Class Purpose:
#   to derive different SVSync from.
#   to allow the user to interact with SVS, fetch and publish.
class SVSyncBase:
    NDN_MTU = 8800
    DATA_INTEREST_LIFETIME = 2000
    DATA_PACKET_FRESHNESS = 10000
    class SVSyncPublicationTooLarge(Exception):
        pass
    def __init__(self, app:NDNApp, syncPrefix:Name, dataPrefix:Name, groupPrefix:Name, nid:Name, updateCallback:Callable, storage:Optional[Storage]=None, securityOptions:Optional[SecurityOptions]=None) -> None:
        SVSyncLogger.info("SVSync: started an svsync type")
        self.app, self.syncPrefix, self.dataPrefix, self.groupPrefix, self.nid, self.updateCallback = app, syncPrefix, dataPrefix, groupPrefix, nid, updateCallback
        self.storage = MemoryStorage() if not storage else storage
        self.secOptions = securityOptions if securityOptions is not None else SecurityOptions(SigningInfo(SignatureType.DIGEST_SHA256), ValidatingInfo(ValidatingInfo.get_validator(SignatureType.DIGEST_SHA256)), SigningInfo(SignatureType.DIGEST_SHA256), [])
        self.app.route(self.dataPrefix)(self.onDataInterest)
        self.core = SVSyncCore(self.app, self.syncPrefix, self.groupPrefix, self.nid, self.updateCallback, self.secOptions)
        SVSyncLogger.info(f'SVSync: started listening to {Name.to_str(self.dataPrefix)}')
    def onDataInterest(self, int_name:FormalName, int_param:InterestParam, _app_param:Optional[BinaryStr]) -> None:
        data_pkt = self.storage.get_packet(int_name, int_param.can_be_prefix)
        if data_pkt:
            SVSyncLogger.info(f'SVSync: served data {Name.to_str(int_name)}')
            self.app.put_raw_packet(data_pkt)
    async def fetchData(self, nid:Name, seqno:int, retries:int=0) -> Optional[bytes]:
        name = self.getDataName(nid, seqno)
        while retries+1 > 0:
            try:
                SVSyncLogger.info(f'SVSync: fetching data {Name.to_str(name)}')
                _, _, _, pkt = await self.app.express_interest(name, need_raw_packet=True, must_be_fresh=True, can_be_prefix=False, lifetime=self.DATA_INTEREST_LIFETIME)
                ex_int_name, meta_info, content, sig_ptrs = parse_data(pkt)
                isValidated = await self.secOptions.validate(ex_int_name, sig_ptrs)
                if not isValidated:
                    return None
                SVSyncLogger.info(f'SVSync: received data {bytes(content)}')
                return bytes(content) if content else None
            except InterestNack as e:
                SVSyncLogger.warning(f'SVSync: nacked with reason={e.reason}')
            except InterestTimeout:
                SVSyncLogger.warning("SVSync: timeout")
            except InterestCanceled:
                SVSyncLogger.warning("SVSync: canceled")
            except ValidationFailure:
                SVSyncLogger.warning("SVSync: data failed to validate")
            except Exception as e:
                SVSyncLogger.warning(f'SVSync: unknown error has occured: {e}')

            retries = retries - 1
            if retries+1 > 0:
                SVSyncLogger.info("SVSync: retrying fetching data")
        return None
    def publishData(self, data:bytes) -> None:
        name = self.getDataName(self.nid, self.core.getSeqno()+1)
        data_packet = make_data(name, MetaInfo(freshness_period=self.DATA_PACKET_FRESHNESS), content=data, signer=self.secOptions.dataSig.signer)
        if len(data_packet) > self.NDN_MTU:
            raise self.SVSyncPublicationTooLarge(f"A SVSync Publication can not be over NDN's MTU ({self.NDN_MTU}).")
        SVSyncLogger.info(f'SVSync: publishing data {Name.to_str(name)}')
        self.storage.put_packet(name, data_packet)
        self.core.updateMyState(self.core.getSeqno()+1)
    def getCore(self) -> SVSyncCore:
        return self.core
    def getDataName(self, nid:Name, seqno:int) -> Name:
        raise NotImplementedError


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from Cryptodome.Hash import SHA256, HMAC
from Cryptodome.PublicKey import ECC, RSA
from Cryptodome.Signature import DSS, pkcs1_15
from typing import Union, Optional, Dict
# NDN Imports
from ndn.encoding import FormalName, SignaturePtrs, Name, SignatureType
from ndn.security import Sha256WithEcdsaSigner, Sha256WithRsaSigner, HmacSha256Signer, DigestSha256Signer, sha256_digest_checker
from ndn.types import Validator
# Custom Imports
from .logger import SVSyncLogger

# Class Type: a struct
# Class Purpose:
#   to hold all signature information including the signer and info used to make the signer
class SigningInfo:
    __slots__ = ('signer','type','keyName','privKey')
    def __init__(self, stype:SignatureType, keyName:Optional[str]=None, privKey:Optional[bytes]=None) -> None:
        self.type, self.signer, self.keyName, self.privKey = stype, None, "", b''

        if self.type != SignatureType.DIGEST_SHA256 and self.type != SignatureType.NOT_SIGNED:
            if privKey is None or privKey == b'':
                raise KeyError("Private Key has to be Defined in Signing Info with this Type.")
            if keyName is None or keyName == "":
                raise KeyError("Key Name has to be Defined in Signing Info with this Type.")
            self.keyName = keyName
            self.privKey = privKey

        if self.type == SignatureType.SHA256_WITH_ECDSA:
            self.signer = Sha256WithEcdsaSigner(Name.from_str(self.keyName), self.privKey)
        elif self.type == SignatureType.SHA256_WITH_RSA:
            self.signer = Sha256WithRsaSigner(Name.from_str(self.keyName), self.privKey)
        elif self.type == SignatureType.HMAC_WITH_SHA256:
            self.signer = HmacSha256Signer(Name.from_str(self.keyName), self.privKey)
        elif self.type == SignatureType.DIGEST_SHA256:
            self.signer = DigestSha256Signer()

# Class Type: a struct with some methods
# Class Purpose:
#   to hold a validator for a specific signature
#   to have a class to generate a basic validator for a certain type
class ValidatingInfo:
    __slots__ = ('validator')
    def __init__(self, validator:Optional[Validator]) -> None:
        self.validator  = validator
    async def validate(self, name:FormalName, sig_ptrs:SignaturePtrs) -> bool:
        if self.validator:
            return await self.validator(name, sig_ptrs)
        return True

    @staticmethod
    def get_validator(stype:SignatureType, keyName:Optional[str]=None, pubKey:Optional[bytes]=None):
        if stype != SignatureType.DIGEST_SHA256 and stype != SignatureType.NOT_SIGNED:
            if pubKey is None or pubKey == b'':
                raise KeyError("Public Key has to be Defined when Generating a Validator with this Type.")
            if keyName is None or keyName == "":
                raise KeyError("Key Name has to be Defined when Generating a Validator with this Type.")
        if stype == SignatureType.SHA256_WITH_ECDSA:
            return ValidatingInfo._ecdsa_checker(Name.from_str(keyName), pubKey)
        if stype == SignatureType.SHA256_WITH_RSA:
            return ValidatingInfo._rsa_checker(Name.from_str(keyName), pubKey)
        if stype == SignatureType.HMAC_WITH_SHA256:
            return ValidatingInfo._hmac_checker(Name.from_str(keyName), pubKey)
        if stype == SignatureType.DIGEST_SHA256:
            return ValidatingInfo._digest_checker()
        return None
    @staticmethod
    def _ecdsa_checker(key_name:FormalName, key_bits:Union[bytes, str]) -> Validator:
        async def wrapper(name:FormalName, sig:SignaturePtrs) -> bool:
            nonlocal key_bits
            sig_info = sig.signature_info
            covered_part = sig.signature_covered_part
            sig_value = sig.signature_value_buf
            if sig_info and sig_info.signature_type == SignatureType.SHA256_WITH_ECDSA:
                if sig_info.key_locator.name != key_name:
                    return False
                pub_key = ECC.import_key(key_bits)
                verifier = DSS.new(pub_key, 'fips-186-3', 'der')
                sha256_algo = SHA256.new()
                if not covered_part or not sig_value:
                    ret = False
                else:
                    for blk in covered_part:
                        sha256_algo.update(blk)
                    try:
                        verifier.verify(sha256_algo, bytes(sig_value))
                        return True
                    except ValueError:
                        return False
                SVSyncLogger.debug(f'Digest check {Name.to_str(name)} -> {ret}')
                return ret
            return False
        return wrapper
    @staticmethod
    def _rsa_checker(key_name:FormalName, key_bits:Union[bytes, str]) -> Validator:
        async def wrapper(name:FormalName, sig:SignaturePtrs) -> bool:
            nonlocal key_bits
            sig_info = sig.signature_info
            covered_part = sig.signature_covered_part
            sig_value = sig.signature_value_buf
            if sig_info and sig_info.signature_type == SignatureType.SHA256_WITH_RSA:
                if sig_info.key_locator.name != key_name:
                    return False
                pub_key = RSA.import_key(key_bits)
                verifier = pkcs1_15.new(pub_key)
                sha256_algo = SHA256.new()
                if not covered_part or not sig_value:
                    ret = False
                else:
                    for blk in covered_part:
                        sha256_algo.update(blk)
                    try:
                        verifier.verify(sha256_algo, bytes(sig_value))
                        return True
                    except ValueError:
                        return False
                SVSyncLogger.debug(f'Digest check {Name.to_str(name)} -> {ret}')
                return ret
            return False
        return wrapper
    @staticmethod
    def _hmac_checker(key_name:FormalName, secret:bytes) -> Validator:
        async def wrapper(name:FormalName, sig:SignaturePtrs) -> bool:
            nonlocal secret
            sig_info = sig.signature_info
            covered_part = sig.signature_covered_part
            sig_value = sig.signature_value_buf
            if sig_info and sig_info.signature_type == SignatureType.HMAC_WITH_SHA256:
                if sig_info.key_locator.name != key_name:
                    return False
                hmac_algo = HMAC.new(secret, digestmod=SHA256)
                if not covered_part or not sig_value:
                    ret = False
                else:
                    for blk in covered_part:
                        hmac_algo.update(blk)
                    try:
                        hmac_algo.verify(sig_value)
                        return True
                    except ValueError:
                        return False
                SVSyncLogger.debug(f'Digest check {Name.to_str(name)} -> {ret}')
                return ret
            return False
        return wrapper
    @staticmethod
    def _digest_checker() -> Validator:
        return sha256_digest_checker

# Class Type: a security class
# Class Purpose:
#   to hold all signing and validating info used in SVS
class SecurityOptions:
    __slots__ = ('syncSig','syncVal','dataSig','dataValDict')
    # It is not neccessary to include digest or nosignature validators in the dataValidatingInfoDict.
    def __init__(self, syncSigningInfo:SigningInfo, syncValidatingInfo:ValidatingInfo, dataSigningInfo:SigningInfo, dataValidatingInfoDict:Dict[str, ValidatingInfo]) -> None:
        self.syncSig, self.syncVal, self.dataSig, self.dataValDict = syncSigningInfo, syncValidatingInfo, dataSigningInfo, dataValidatingInfoDict
    async def validate(self, name:FormalName, sig_ptrs:SignaturePtrs):
        val = None
        if sig_ptrs.signature_info.signature_type is None:
            return True
        if sig_ptrs.signature_info.signature_type == 1:
            val = ValidatingInfo(ValidatingInfo.get_validator(SignatureType.DIGEST_SHA256))
        else:
            try:
                keyname = Name.to_str(sig_ptrs.signature_info.key_locator.name)
                for key in self.dataValDict:
                    if key == keyname:
                        val = self.dataValDict[keyname]
            except AttributeError:
                val = None
        if val:
            return await val.validate(name, sig_ptrs)
        return True # We do not have the key for this keyname (cant error check it)


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# API classes
from .svs_base import SVSyncBase
from .svs import SVSync
from .svs_shared import SVSyncShared

# Forked Thread API classes
from .svs_base_thread import SVSyncBase_Thread
from .svs_thread import SVSync_Thread
from .svs_shared_thread import SVSyncShared_Thread

# Logic of SVS
from .core import SVSyncCore
from .balancer import SVSyncBalancer

# Additional SVS useful classes
from .meta_data import MetaData
from .missing_data import MissingData
from .state_vector import StateVector
from .state_table import StateTable
from .security import SigningInfo, ValidatingInfo, SecurityOptions
from .tlv import SVSyncTlvTypes

# Tool classes
from .scheduler import AsyncScheduler
from .window import AsyncWindow

# Logging class
from .logger import SVSyncLogger


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
import logging
from sys import stdout
from typing import Optional

# Class Type: a logging class
# Class Purpose:
#   to properly handle all logging within this library
class SVSyncLogger(object):
    LOGGER_NAME = "ndn-svs"
    LOGGER_LEVEL = logging.DEBUG
    LOGGER_FORMAT = "(%(asctime)s) %(name)s:%(levelname)s | %(message)s"
    CONSOLE_LOGGING = False
    LOGGER_FILE = None
    LOGGER = None
    CONFIGURED = False

    @staticmethod
    def config(console:bool, file:Optional[str], level:int, lformat:Optional[str]=None) -> None:
        if not SVSyncLogger.CONFIGURED:
            SVSyncLogger.CONFIGURED = True
            SVSyncLogger.LOGGER_LEVEL = level
            SVSyncLogger.CONSOLE_LOGGING = console
            SVSyncLogger.LOGGER_FILE = file
            SVSyncLogger.LOGGER_FORMAT = lformat if lformat else SVSyncLogger.LOGGER_FORMAT

            SVSyncLogger.LOGGER = logging.getLogger(SVSyncLogger.LOGGER_NAME)
            SVSyncLogger.LOGGER.setLevel(SVSyncLogger.LOGGER_LEVEL)
            fm = logging.Formatter(SVSyncLogger.LOGGER_FORMAT)
            if SVSyncLogger.CONSOLE_LOGGING:
                console_handler = logging.StreamHandler(stdout)
                console_handler.setFormatter(fm)
                SVSyncLogger.LOGGER.addHandler(console_handler)
            if SVSyncLogger.LOGGER_FILE:
                file_handler = logging.FileHandler(SVSyncLogger.LOGGER_FILE)
                file_handler.setFormatter(fm)
                SVSyncLogger.LOGGER.addHandler(file_handler)
            SVSyncLogger.LOGGER.propagate = False

    @staticmethod
    def debug(msg:str) -> None:
        if SVSyncLogger.CONFIGURED:
            SVSyncLogger.LOGGER.debug(msg)
    @staticmethod
    def info(msg:str) -> None:
        if SVSyncLogger.CONFIGURED:
            SVSyncLogger.LOGGER.info(msg)
    @staticmethod
    def warning(msg:str) -> None:
        if SVSyncLogger.CONFIGURED:
            SVSyncLogger.LOGGER.warning(msg)
    @staticmethod
    def error(msg:str) -> None:
        if SVSyncLogger.CONFIGURED:
            SVSyncLogger.LOGGER.error(msg)
    @staticmethod
    def critical(msg:str) -> None:
        if SVSyncLogger.CONFIGURED:
            SVSyncLogger.LOGGER.critical(msg)


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
from typing import List, Optional
# NDN Imports
from ndn.encoding import Component, Name
# Custom Imports
from .meta_data import MetaData
from .missing_data import MissingData
from .state_vector import StateVector, StateVectorEntry

# Class Type: a class
# Class Purpose:
#   to allow an easier time to form sync interest and handle all states
class StateTable:
    NDN_MTU = 8800
    def __init__(self, mynid:Name) -> None:
        self.table, self.meta, self.parts, self.nidstr = StateVector(), MetaData(), [[0,0]], Name.to_str(mynid)
        self.meta.source = self.nidstr.encode()
    def processStateVector(self, incoming_sv:StateVector, oldData:bool) -> List[MissingData]:
        listOrder:List[StateVectorEntry] = incoming_sv.list() if oldData else reversed(incoming_sv.list())
        missingList:List[MissingData] = []
        for i in listOrder:
            tableSeqno:int = self.table.get(i.nid) if self.table.get(i.nid) else 0
            if tableSeqno < i.seqno:
                tableSeqno = tableSeqno + 1
                temp:MissingData = MissingData(i.nid, tableSeqno, i.seqno)
                self.table.set(i.nid, i.seqno, oldData)
                missingList.append(temp)
        return missingList
    def updateMetaData(self) -> None:
        self.meta.tseqno = self.table.total()
        self.calculateParts()
        self.meta.nopcks = len(self.parts) - 1
    def updateMyState(self, seqno:int) -> None:
        self.table.set(self.nidstr, seqno)
    def getPart(self, no:int) -> Component:
        try:
            return self.table.partition(self.parts[no][0],self.parts[no][1])
        except (IndexError, ValueError):
            return bytearray()
    def getMetaData(self) -> MetaData:
        return self.meta
    def getCompleteStateVector(self) -> StateVector:
        return self.table
    def calculateParts(self) -> None:
        mainlist, templist, part, total = [], [], 0, 0
        templist.insert(0, 0)
        for index, value in enumerate(self.table.entry_lengths()):
            if total + value > self.getPartMaximum(part):
                templist.insert(1, index)
                part = part + 1
                total = value
                mainlist.append(templist)
                templist = []
                templist.insert(0, index)
            else:
                total = total + value
        if templist:
            templist.insert(1, len(self.table.entry_lengths()))
            mainlist.append(templist)
        self.parts = mainlist
    def getPartMaximum(self, part:int) -> int:
        return (self.NDN_MTU - 1500) # optimize this part for each packet (including interest)
    def getPartCuts(self) -> list:
        return self.parts
    def getSeqno(self, nid:Name) -> Optional[int]:
        return self.table.get(Name.to_str(nid))


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Class Type: a struct
# Class Purpose:
#   to hold the range of missing data for a specific node.
class MissingData:
    __slots__ = ('nid','lowSeqno','highSeqno')
    def __init__(self, nid:str, lowSeqno:int, highSeqno:int) -> None:
        self.nid, self.lowSeqno, self.highSeqno = nid, lowSeqno, highSeqno


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
import asyncio as aio
from typing import Optional, Callable
# NDN Imports
from ndn.app import NDNApp
from ndn.encoding import Name, Component, InterestParam, BinaryStr, FormalName, SignaturePtrs, parse_data
from ndn.types import InterestNack, InterestTimeout, InterestCanceled, ValidationFailure
# Custom Imports
from .logger import SVSyncLogger
from .meta_data import MetaData
from .security import SecurityOptions
from .state_table import StateTable
from .state_vector import StateVector
from .window import AsyncWindow

# Class Type: a ndn class
# Class Purpose:
#   manage sync interests that are sent out.
#   to hear about other sync interests
#   to find out about new data from other nodes.
class SVSyncBalancer:
    PROP_WINDOW = 5
    PROP_INTEREST_LIFETIME = 2000
    PROP_PACKET_FRESHNESS = 5000
    def __init__(self, app:NDNApp, groupPrefix:Name, nid:Name, table:StateTable, updateCallback:Callable, secOptions:SecurityOptions) -> None:
        SVSyncLogger.info("SVSyncBalancer: started svsync balancer")
        self.app, self.groupPrefix, self.nid, self.table, self.updateCallback, self.secOptions, self.busy = app, groupPrefix, nid, table, updateCallback, secOptions, False
        self.propPrefix = self.nid + self.groupPrefix + [Component.from_str("prop")]
        self.taskWindow = AsyncWindow(self.PROP_WINDOW)
        self.app.route(self.propPrefix, need_sig_ptrs=True)(self.onPropInterest)
        SVSyncLogger.info(f'SVSyncBalancer: started listening to {Name.to_str(self.propPrefix)}')
    async def balanceFromProp(self, name:Name, pckno:int) -> None:
        incoming_sv:Optional[StateVector] = await self.sendPropInterest(name, pckno)
        if not incoming_sv:
            return
        missingList = self.table.processStateVector(incoming_sv, oldData=True)
        if missingList:
            self.updateCallback(missingList)
        self.table.updateMetaData()
    async def equalize(self, incoming_md:MetaData) -> None:
        if incoming_md.tseqno <= self.table.getMetaData().tseqno or self.busy:
            return
        self.busy = True
        name = Name.from_str(bytes(incoming_md.source).decode())
        for i in range(incoming_md.nopcks):
            self.taskWindow.addTask(self.balanceFromProp, (name, i+1))
        await self.taskWindow.gather()
        SVSyncLogger.info(f'SVSyncBalancer: nmeta {bytes(self.table.getMetaData().source).decode()} - {self.table.getMetaData().tseqno} total, {self.table.getMetaData().nopcks} pcks')
        SVSyncLogger.info(f'SVSyncBalancer: ntable {self.table.getCompleteStateVector().to_str()}')
        self.busy = False
    def onPropInterest(self, int_name:FormalName, int_param:InterestParam, _app_param:Optional[BinaryStr], sig_ptrs:SignaturePtrs) -> None:
        SVSyncLogger.info(f'SVSyncBalancer: received balance {Name.to_str(int_name)}')
        aio.get_event_loop().create_task(self.onPropInterestHelper(int_name, int_param, _app_param, sig_ptrs))
    async def onPropInterestHelper(self, int_name:FormalName, int_param:InterestParam, _app_param:Optional[BinaryStr], sig_ptrs:SignaturePtrs) -> None:
        sv = bytes(self.table.getPart(Component.to_number(int_name[-1])))
        SVSyncLogger.info(f'SVSyncBalancer: sending balance {sv}')
        self.app.put_data(int_name, content=sv, signer=self.secOptions.dataSig.signer, freshness_period=self.PROP_PACKET_FRESHNESS)
    async def sendPropInterest(self, source:Name, pckno:int) -> Optional[StateVector]:
        name:Name = source + self.groupPrefix + [Component.from_str("prop")] + [Component.from_number(pckno, Component.TYPE_SEQUENCE_NUM)]
        try:
            SVSyncLogger.info(f'SVSyncBalancer: balancing by {Name.to_str(name)}')
            _, _, _, pkt = await self.app.express_interest(
                name, need_raw_packet=True, must_be_fresh=True, can_be_prefix=False, lifetime=self.PROP_INTEREST_LIFETIME)
            int_name, meta_info, content, sig_ptrs = parse_data(pkt)
            isValidated = await self.secOptions.validate(int_name, sig_ptrs)
            return StateVector(bytes(content)) if bytes(content) != b'' and isValidated else None
        except (InterestNack, InterestTimeout, InterestCanceled, ValidationFailure):
            SVSyncLogger.info(f'SVSyncBalancer: failed to balance from {Name.to_str(name)}')
            return None
    def isBusy(self) -> bool:
        return self.busy


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

# Basic Libraries
import asyncio as aio
from enum import Enum
from random import uniform
from typing import Callable, Optional, Tuple, List
# NDN Imports
from ndn.app import NDNApp
from ndn.encoding import Name, InterestParam, BinaryStr, FormalName, SignaturePtrs
from ndn.types import InterestNack, InterestTimeout, InterestCanceled, ValidationFailure
# Custom Imports
from .balancer import SVSyncBalancer
from .state_table import StateTable
from .meta_data import MetaData
from .state_vector import StateVector
from .scheduler import AsyncScheduler
from .security import SecurityOptions
from .logger import SVSyncLogger
from .missing_data import MissingData

# Class Type: an enumeration struct
# Class Purpose:
#   to differ core states.
class SVSyncCoreState(Enum):
    STEADY     = 0
    SUPRESSION = 1

# Class Type: a ndn class
# Class Purpose:
#   manage sync interests that are sent out.
#   to hear about other sync interests
#   to find out about new data from other nodes.
class SVSyncCore:
    INTERVAL = 30000
    INTERVAL_RANDOMNESS = 0.1
    BRIEF_INTERVAL = 200
    BRIEF_INTERVAL_RANDOMNESS = 0.5
    SYNC_INTEREST_LIFETIME = 1000
    def __init__(self, app:NDNApp, syncPrefix:Name, groupPrefix:Name, nid:Name, updateCallback:Callable, secOptions:SecurityOptions) -> None:
        SVSyncLogger.info("SVSyncCore: started svsync core")
        self.app, self.nid, self.updateCallback, self.syncPrefix, self.groupPrefix, self.secOptions, self.state = app, nid, updateCallback, syncPrefix, groupPrefix, secOptions, SVSyncCoreState.STEADY
        self.table = StateTable(self.nid)
        self.balancer = SVSyncBalancer(self.app, self.groupPrefix, self.nid, self.table, self.updateCallback, self.secOptions)
        self.app.route(self.syncPrefix, need_sig_ptrs=True)(self.onSyncInterest)
        SVSyncLogger.info(f'SVSyncCore: started listening to {Name.to_str(self.syncPrefix)}')
        self.scheduler = AsyncScheduler(self.sendSyncInterest, self.INTERVAL, self.INTERVAL_RANDOMNESS)
        self.scheduler.skip_interval()
    async def asyncSendSyncInterest(self) -> None:
        name:Name = self.syncPrefix + [self.table.getMetaData().encode()] + [ self.table.getPart(0) ]
        SVSyncLogger.info(f'SVSyncCore: sync {Name.to_str(name)}')
        try:
            data_name, meta_info, content = await self.app.express_interest(
                name, signer=self.secOptions.syncSig.signer, must_be_fresh=True, can_be_prefix=True, lifetime=self.SYNC_INTEREST_LIFETIME)
        except (InterestNack, InterestTimeout, InterestCanceled, ValidationFailure):
            pass
    def sendSyncInterest(self) -> None:
        aio.get_event_loop().create_task(self.asyncSendSyncInterest())
    def onSyncInterest(self, int_name:FormalName, int_param:InterestParam, _app_param:Optional[BinaryStr], sig_ptrs:SignaturePtrs) -> None:
        aio.get_event_loop().create_task(self.onSyncInterestHelper(int_name, int_param, _app_param, sig_ptrs))
    async def onSyncInterestHelper(self, int_name:FormalName, int_param:InterestParam, _app_param:Optional[BinaryStr], sig_ptrs:SignaturePtrs) -> None:
        isValidated:bool = await self.secOptions.syncVal.validate(int_name, sig_ptrs)
        if not isValidated:
            return

        incomingVector, incomingMetadata = StateVector(int_name[-2]), MetaData(int_name[-3])
        SVSyncLogger.info("SVSyncCore: >> I: received sync")
        SVSyncLogger.info(f'SVSyncCore:       rmeta {bytes(incomingMetadata.source).decode()} - {incomingMetadata.tseqno} total, {incomingMetadata.nopcks} pcks')
        SVSyncLogger.info(f'SVSyncCore:       {incomingVector.to_str()}')

        missingList:List[MissingData] = self.table.processStateVector(incomingVector, oldData=False)
        self.table.updateMetaData()
        if missingList:
            self.updateCallback(missingList)

        supress, equalize = self.compareMetaData(incomingMetadata)
        self.state = SVSyncCoreState.SUPRESSION if supress else SVSyncCoreState.STEADY

        # reset the sync timer if STEADY
        # supress the timer if SUPPRESION
        if self.state == SVSyncCoreState.STEADY:
            self.scheduler.set_cycle()
        else:
            delay = self.BRIEF_INTERVAL + round( uniform(-self.BRIEF_INTERVAL_RANDOMNESS,self.BRIEF_INTERVAL_RANDOMNESS)*self.BRIEF_INTERVAL )
            if self.scheduler.get_time_left() > delay:
                self.scheduler.set_cycle(delay)
        SVSyncLogger.info(f'SVSyncCore: state {self.state.name}')
        SVSyncLogger.info(f'SVSyncCore: parts-{self.table.getPartCuts()} | 1stlength-{len(self.table.getPart(0))}')
        SVSyncLogger.info(f'SVSyncCore: table {self.table.getCompleteStateVector().to_str()}')

        if equalize:
            await self.balancer.equalize(incomingMetadata)
    def compareMetaData(self, incoming_md:MetaData) -> Tuple[bool, bool]:
        table_md, supress, equalize = self.table.getMetaData(), False, False
        if table_md.tseqno > incoming_md.tseqno:
            supress = True
        elif table_md.tseqno < incoming_md.tseqno:
            equalize = True
        return (supress, equalize)
    def updateMyState(self, seqno:int) -> None:
        self.table.updateMyState(seqno)
        self.table.updateMetaData()
        self.scheduler.skip_interval()
    def getSeqno(self) -> int:
        seqno = self.table.getSeqno(self.nid)
        return seqno if seqno != None else 0
    def getStateTable(self) -> StateTable:
        return self.table


#    @Author: Justin C Presley
#    @Author-Email: justincpresley@gmail.com
#    @Project: NDN State Vector Sync Protocol
#    @Source-Code: https://github.com/justincpresley/ndn-python-svs
#    @Pip-Library: https://pypi.org/project/ndn-svs
#    @Documentation: https://ndn-python-svs.readthedocs.io

#!/usr/bin/env python3
import io
import re
from setuptools import setup, find_packages
from typing import List

with io.open("docs/version.py", "rt", encoding="utf8") as f:
    version = re.search(r'__version__ = "(.*?)"', f.read()).group(1)

with io.open("README.rst", "rt", encoding="utf8") as f:
    long_description = f.read()

def _parse_requirements(filename: str) -> List[str]:
    with open(filename, 'r') as f:
        return [s for s in [line.split('#', 1)[0].strip(' \t\n') for line in f ] if s != '']

setup(
    name='ndn-svs',
    version=version,
    description='The NDN State Vector Sync (SVS) Protocol in Python 3',
    long_description=long_description,
    long_description_content_type='text/x-rst',
    url='https://github.com/justincpresley/ndn-python-svs',
    author='Justin C Presley',
    author_email='justincpresley@gmail.com',
    maintainer='Justin C Presley',
    maintainer_email='justincpresley@gmail.com',
    download_url='https://pypi.python.org/pypi/ndn-svs',
    project_urls={
        "Bug Tracker": "https://github.com/justincpresley/ndn-python-svs/issues",
        "Source Code": "https://github.com/justincpresley/ndn-python-svs",
    },
    license='LGPL-2.1 License',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Libraries',
        'Topic :: Internet',
        'Topic :: System :: Networking',
        'License :: OSI Approved :: GNU Lesser General Public License v2 (LGPLv2)',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10'
    ],
    keywords='NDN SVS',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=_parse_requirements('docs/requirements.txt'),
    python_requires=">=3.8",
    zip_safe=False)